@@ -1,10 +1,12 @@
 (ns ring.middleware.content-type
   "Middleware for automatically adding a content type to response maps."
-  (:use ring.util.mime-type
-        ring.util.response))
+  (:require [ring.util.mime-type :refer [ext-mime-type]]
+            [ring.util.response :refer [content-type get-header]]))
 
 (defn content-type-response
-  "Adds a content-type header to response."
+  "Adds a content-type header to response. See: wrap-content-type."
+  {:arglists '([response request] [response request options])
+   :added "1.2"}
   [resp req & [opts]]
   (if (get-header resp "Content-Type")
     resp
@@ -18,9 +20,11 @@
   content-type can be found, it defaults to 'application/octet-stream'.
 
   Accepts the following options:
-    :mime-types - a map of filename extensions to mime-types that will be
-                  used in addition to the ones defined in
-                  ring.util.mime-types/default-mime-types"
+
+  :mime-types - a map of filename extensions to mime-types that will be
+                used in addition to the ones defined in
+                ring.util.mime-types/default-mime-types"
+  {:arglists '([handler] [handler options])}
   [handler & [opts]]
   (fn [req]
     (if-let [resp (handler req)]
