@@ -5,7 +5,8 @@
             [puppetlabs.puppetdb.random :refer :all]
             [puppetlabs.puppetdb.reports :as reports]
             [puppetlabs.puppetdb.schema :as pls]
-            [schema.core :as s]))
+            [schema.core :as s])
+  (:import [org.apache.commons.lang3 StringUtils]))
 
 ;; Validation functions, for use within pre/post conditions
 
@@ -109,6 +110,12 @@
       (rule-match? x context) (get x "anonymize")
       :else (recur (rest xs)))))
 
+(defn anonymize-text
+  "This is for anonymizing text data where we care only about size, not content."
+  [text]
+  (when text
+    (StringUtils/repeat "?" (count text))))
+
 ;; Functions for anonymizing the final leaf data
 (defn anonymize-leaf-value
   "Based on the input value, return an appropriate random replacement"
@@ -132,8 +139,8 @@
        :type (random-type-name)
        :title (random-string (max 10 (count value)))
        :parameter-name (random-string-alpha (max 10 (count value)))
+       :text (anonymize-text value)
        :message (random-string (max 10 (count value)))
-       :log-message (random-string (max 10 (count value)))
        :file (random-pp-path)
        :line (when value (rand-int (max value 20)))
        :transaction_uuid (uuid)
@@ -369,7 +376,7 @@
 
 (defn anonymize-log [log context config]
   (-> log
-      (update "message" anonymize-leaf :log-message context config)
+      (update "message" anonymize-leaf :text context config)
       (update "source" anonymize-log-source context config)
       (update "tags" anonymize-tags context config)
       (update "file" anonymize-leaf :file context config)
@@ -574,7 +581,7 @@
    "none" {
            "rules" {
                     "node" [ {"context" {} "anonymize" false} ]
-                    "log-message" [ {"context" {} "anonymize" false} ]
+                    "text" [ {"context" {} "anonymize" false} ]
                     "type" [ {"context" {} "anonymize" false} ]
                     "title" [ {"context" {} "anonymize" false} ]
                     "parameter-name" [ {"context" {} "anonymize" false} ]
