@@ -51,8 +51,13 @@
             [puppetlabs.puppetdb.random :refer [random-string random-bool]]
             [puppetlabs.puppetdb.archive :as archive]
             [slingshot.slingshot :refer [try+ throw+]]
-            [clojure.core.async :refer [go go-loop >! <! >!! <!! timeout chan alt! close! dropping-buffer]]
-            [clojure.core.match :as cm]))
+            [clojure.core.async :refer [go go-loop >! <! >!! <!! timeout chan close! dropping-buffer pipeline]]
+            [clojure.core.match :as cm]
+            [puppetlabs.puppetdb.mq :as mq]
+            [taoensso.nippy :as nippy])
+  (:import [org.apache.activemq.broker BrokerService]
+           [javax.jms MessageConsumer MessageProducer Session]
+           [clojure.core.async.impl.protocols Buffer]))
 
 (def cli-description "Development-only benchmarking tool")
 
@@ -196,10 +201,8 @@
 
 (defn update-host
   "Submit a `catalog` for `hosts` (when present), possibly mutating it before
-   submission.  Also submit a report for the host (if present). This is
-   similar to timed-update-host, but always sends the update (doesn't run/skip
-   based on the clock)"
-  [{:keys [catalog report factset] :as state} rand-percentage current-time command-send-ch]
+  submission. Also submit a report for the host (if present)."
+  [{:keys [host catalog report factset] :as state} rand-percentage current-time command-send-ch]
   (let [stamp (jitter current-time 1800)
         uuid (kitchensink/uuid)
         catalog (some-> catalog
@@ -208,35 +211,14 @@
                        (update-report uuid stamp))
         factset (some-> factset
                         (update-factset rand-percentage stamp))]
-    (when catalog (>!! command-send-ch [:catalog 6 catalog]))
-    (when report (>!! command-send-ch [:report 6 report]))
-    (when factset (>!! command-send-ch [:factset 4 factset]))
+    (when catalog (>!! command-send-ch [:catalog host 6 catalog]))
+    (when report (>!! command-send-ch [:report host 6 report]))
+    (when factset (>!! command-send-ch [:factset host 4 factset]))
 
     (assoc state
            :catalog catalog
            :factset factset)))
 
-(defn run-simulated-hosts
-  "Run a host simulation job."
-  [hosts run-interval-min rand-percentage command-send-ch]
-  (let [close-to-stop-ch (chan)
-        host-queue (into clojure.lang.PersistentQueue/EMPTY hosts)
-        ms-per-host (float (/ (* 60000 run-interval-min) (count host-queue)))]
-    (go
-      (alt!
-        close-to-stop-ch
-        ::stopped
-
-        (timeout (rand 1000)) ; random initial offset from other simulation threads
-        (go-loop [host-queue host-queue]
-          (let [host-timeout (timeout ms-per-host)
-                updated-host (update-host (first host-queue) rand-percentage (time/now) command-send-ch)
-                new-host-queue (conj (pop host-queue) updated-host)]
-            (alt!
-              close-to-stop-ch ::stopped
-              host-timeout (recur new-host-queue))))))
-    close-to-stop-ch))
-
 (defn submit-n-messages
   "Given a list of host maps, send `num-messages` to each host.  The function
    is recursive to accumulate possible catalog mutations (i.e. changing a previously
@@ -328,18 +310,19 @@
 
 (defn try-read-and-send-command [base-url command-send-ch]
   (try
-   (cm/match [(<!! command-send-ch)]
-             [:stop] nil
-             [nil] nil
-             [[entity version payload]] (do ((case entity
-                                               :catalog client/submit-catalog
-                                               :report client/submit-report
-                                               :factset client/submit-facts)
-                                             base-url version payload)
-                                            ::submitted))
-   (catch Exception e
-     (println "Exception while submitting command: " e)
-     ::error)))
+    (cm/match [(<!! command-send-ch)]
+              [:stop] nil
+              [nil] nil
+              [[entity host version payload]]
+              (do ((case entity
+                     :catalog client/submit-catalog
+                     :report client/submit-report
+                     :factset client/submit-facts)
+                   base-url host version payload)
+                  ::submitted))
+    (catch Exception e
+      (println "Exception while submitting command: " e)
+      ::error)))
 
 (defn start-command-sender
   "Start a command sending thread. Reads commands from command-send-ch of the
@@ -389,6 +372,83 @@
                    (cons (first s) (first buckets))))
       (seq buckets))))
 
+(defn make-jms-nippy-message
+  "Create a jms message containing the clojure data structure 'payload' encoded
+  into a byte array using nippy."
+  [session payload]
+  (let [msg (.createBytesMessage session)
+        payload-bytes (nippy/freeze payload)]
+    (.writeBytes msg payload-bytes)
+    msg))
+
+(defn unpack-jms-nippy-message
+  "Return the clojure data payload from a message created with
+  'make-jms-nippy-message'."
+  [msg]
+  (let [bytes (byte-array (.getBodyLength msg))]
+    (.readBytes msg bytes)
+    (nippy/thaw bytes)))
+
+(deftype AMQBrokerBuffer [^BrokerService broker
+                          ^Session session
+                          ^MessageProducer producer
+                          ^MessageConsumer consumer
+                          mq-dir
+                          queue-count]
+  Buffer
+  (full? [this] false)
+  (remove! [this]
+    (swap! queue-count dec)
+    (mq/commit-or-rollback session (unpack-jms-nippy-message (.receive consumer))))
+  (add!* [this item]
+    (swap! queue-count inc)
+    (mq/commit-or-rollback session (.send producer (make-jms-nippy-message session item))))
+  (close-buf! [this]
+    (println "Shutting down AMQ...")
+    (reset! queue-count 0)
+    (mq/stop-broker! broker)
+    (fs/delete-dir mq-dir))
+  clojure.lang.Counted
+  (count [this] @queue-count))
+
+(defn amq-broker-buffer [broker-name endpoint-name]
+  (let [dir (fs/absolute-path broker-name)
+        conn-str (str "vm://" broker-name
+                      "?broker.useJmx=false"
+                      "&broker.enableStatistics=false"
+                      "&jms.useCompression=true"
+                      "&jms.copyMessageOnSend=false")
+        size-megs 20000
+        broker (mq/build-embedded-broker broker-name
+                                         dir
+                                         {:store-usage size-megs
+                                          :temp-usage  size-megs})
+        _ (println "Using amq in" dir)
+        _ (.setPersistent broker false)
+        _ (mq/start-broker! broker)
+        factory (mq/activemq-connection-factory conn-str)
+        connection (doto (.createConnection factory) .start)
+        session (.createSession connection true 0)]
+    (AMQBrokerBuffer. broker
+                      session
+                      (.createProducer session (.createQueue session endpoint-name))
+                      (.createConsumer session (.createQueue session endpoint-name))
+                      dir
+                      (atom 0))))
+
+(defn relay-messages-at-rate
+  "Relay messages from in-ch to out-ch at the given rate in
+  messages-per-second."
+  [in-ch out-ch messages-per-second]
+  (let [ms-per-message (- (/ 1000 messages-per-second) 3)]
+    (go-loop []
+      (let [message (<! in-ch)
+            message-timeout (timeout (int (+ (rand) ms-per-message)))]
+        (when message
+          (>! out-ch message)
+          (<! message-timeout)
+          (recur))))))
+
 (defn benchmark-main
   [& args]
   (let [{:keys [config rand-perc numhosts nummsgs threads] :as options} (validate-cli! args)
@@ -415,7 +475,7 @@
                           (chan (dropping-buffer 10000)))
         rate-monitor-ch (chan (* 2 threads))
         run-interval (-> (get options :runinterval 30) time/minutes)
-        simulation-threads 10
+        simulation-threads 4
         commands-per-puppet-run (+ (if catalogs 1 0)
                                    (if reports 1 0)
                                    (if facts 1 0))]
@@ -436,30 +496,41 @@
           (>!! command-send-ch :stop))
         (close! command-send-ch))
 
-      (let [run-interval-minutes (time/in-minutes run-interval)
-            partitioned-hosts (->> hosts
-                                   (map #(assoc % :lastrun (rand-lastrun run-interval)))
-                                   (partition-into-buckets simulation-threads))
-            close-to-stop-chans (mapv (fn [hosts-for-partition]
-                                        (run-simulated-hosts hosts-for-partition
-                                                             run-interval-minutes
-                                                             rand-perc
-                                                             command-send-ch))
-                                      partitioned-hosts)
-            close-to-stop-ch (chan)]
-        (go
-          (<! close-to-stop-ch)
-          (close! rate-monitor-ch)
-          (close! command-send-ch)
-          (dorun (map close! close-to-stop-chans)))
+      (let [close-to-stop-ch (chan)
+            mq (chan (amq-broker-buffer "benchmark-mq" "benchmark-endpoint"))]
 
+        ;; be sure to clear the temp queue on exit
+        (.addShutdownHook (Runtime/getRuntime) (Thread. #(close! mq)))
+
+        ;; initial queue population
+        (future
+          (println "Populating the queue in the background")
+          (dotimes [n numhosts]
+            (let [h (-> (make-host n)
+                        (assoc :lastrun (rand-lastrun run-interval)))]
+              (>!! mq h)))
+
+          (println "... finished filling the queue, continuing with the simulation."))
+
+        (let [run-interval-minutes (time/in-minutes run-interval)
+              hosts-per-second (/ numhosts (* run-interval-minutes 60))
+              rate-limited-mq-out (chan)]
+          (relay-messages-at-rate mq rate-limited-mq-out hosts-per-second)
+          (pipeline simulation-threads
+                    mq
+                    (map #(update-host % rand-perc (time/now) command-send-ch))
+                    rate-limited-mq-out)
+          (go
+            (<! close-to-stop-ch)
+            (mapv close! [rate-limited-mq-out mq rate-monitor-ch command-send-ch])))
         close-to-stop-ch))))
 
 (defn chan? [x]
   (satisfies? clojure.core.async.impl.protocols/Channel x))
 
+
 (defn -main [& args]
   (let [x (apply benchmark-main args)]
-   (when (chan? x)
-     (println "Press ctrl-c to stop.")
-     (<!! x))))
+    (when (chan? x)
+      (println "Press ctrl-c to stop.")
+      (<!! x))))
