@@ -44,8 +44,10 @@
      maintain acceptable performance."
   (:require [clj-time.core :refer [ago]]
             [clojure.java.io :as io]
+            [clojure.java.jmx :as jmx]
             [clojure.tools.logging :as log]
             [compojure.core :as compojure]
+            [metrics.reporters.jmx :as jmx-reporter]
             [overtone.at-at :refer [mk-pool interspaced]]
             [puppetlabs.kitchensink.core :as kitchensink]
             [puppetlabs.puppetdb.cheshire :as json]
@@ -55,12 +57,13 @@
             [puppetlabs.puppetdb.http.server :as server]
             [puppetlabs.puppetdb.jdbc :as jdbc]
             [puppetlabs.puppetdb.meta.version :as version]
-            [puppetlabs.puppetdb.scf.storage-utils :as sutils]
+            [puppetlabs.puppetdb.metrics.core :as metrics]
             [puppetlabs.puppetdb.mq :as mq]
             [puppetlabs.puppetdb.query-eng :as qeng]
             [puppetlabs.puppetdb.query.population :as pop]
             [puppetlabs.puppetdb.scf.migrate :refer [migrate! indexes!]]
             [puppetlabs.puppetdb.scf.storage :as scf-store]
+            [puppetlabs.puppetdb.scf.storage-utils :as sutils]
             [puppetlabs.puppetdb.time :refer [to-seconds to-millis parse-period
                                               format-period period?]]
             [puppetlabs.puppetdb.utils :as utils]
@@ -71,6 +74,7 @@
   (:import [javax.jms ExceptionListener]))
 
 (def cli-description "Main PuppetDB daemon")
+(def database-metrics-registry (get-in metrics/metrics-registries [:database :registry]))
 
 ;; ## Wiring
 ;;
@@ -191,16 +195,16 @@
 
 (defn initialize-schema
   "Ensure the database is migrated to the latest version, and warn if
-  it's deprecated, log and exit if it's unsupported. We do this in a
-  single connection because HSQLDB seems to get confused if the
-  database doesn't exist but we open and close a connection without
-  creating anything."
+  it's deprecated, log and exit if it's unsupported.
+
+  This function will return true iff any migrations were run."
   [db-conn-pool config]
   (jdbc/with-db-connection db-conn-pool
     (scf-store/validate-database-version #(System/exit 1))
     @sutils/db-metadata
-    (migrate! db-conn-pool)
-    (indexes! config)))
+    (let [migrated? (migrate! db-conn-pool)]
+      (indexes! config)
+      migrated?)))
 
 (defn init-with-db
   "All initialization operations needing a database connection should
@@ -208,16 +212,25 @@
   `write-db-config` that will hang until it is able to make a
   connection to the database. This covers the case of the database not
   being fully started when PuppetDB starts. This connection pool will
-  be opened and closed within the body of this function."
+  be opened and closed within the body of this function.
+
+  This function will return true iff any migrations were run."
   [write-db-config config]
-  (with-open [init-db-pool (jdbc/make-connection-pool
-                            (assoc write-db-config
-                                   ;; Block waiting to grab a connection
-                                   :connection-timeout 0
-                                   ;; Only allocate connections when needed
-                                   :pool-availability-threshold 0))]
-    (let [db-pool-map {:datasource init-db-pool}]
-      (initialize-schema db-pool-map config))))
+  (loop [db-spec (assoc write-db-config
+                        ;; Block waiting to grab a connection
+                        :connection-timeout 15000
+                        :pool-name "PDBMigrationsPool")]
+    (if-let [result
+             (try
+               (with-open [init-db-pool (jdbc/make-connection-pool db-spec
+                                                                   database-metrics-registry)]
+                 (let [db-pool-map {:datasource init-db-pool}]
+                   (initialize-schema db-pool-map config)
+                   ::success))
+               (catch java.sql.SQLTransientConnectionException e
+                 (log/error e "Error while attempting to create connection pool")))]
+      result
+      (recur db-spec))))
 
 (defn start-puppetdb
   [context config service get-registered-endpoints]
@@ -225,24 +238,29 @@
          (map? config)]
    :post [(map? %)
           (every? (partial contains? %) [:broker])]}
-  (let [{:keys [global   jetty
+  (let [{:keys [developer jetty
                 database read-database
                 puppetdb command-processing]} config
-        {:keys [vardir catalog-hash-debug-dir]} global
+        {:keys [pretty-print]} developer
         {:keys [gc-interval dlo-compression-interval node-ttl
                 node-purge-ttl report-ttl]} database
         {:keys [dlo-compression-threshold]} command-processing
         {:keys [disable-update-checking]} puppetdb
 
-        write-db (jdbc/pooled-datasource database)
-        read-db (jdbc/pooled-datasource (assoc read-database :read-only? true))
+        write-db (jdbc/pooled-datasource (assoc database :pool-name "PDBWritePool")
+                                         database-metrics-registry)
+        read-db (jdbc/pooled-datasource (assoc read-database :read-only? true :pool-name "PDBReadPool")
+                                        database-metrics-registry)
         discard-dir (io/file (conf/mq-discard-dir config))]
 
     (when-let [v (version/version)]
       (log/infof "PuppetDB version %s" v))
 
-    (init-with-db database config)
-    (pop/initialize-metrics read-db)
+    (when (init-with-db database config)
+      (sutils/async-vacuum-analyze write-db))
+
+    (let [population-registry (get-in metrics/metrics-registries [:population :registry])]
+      (pop/initialize-population-metrics! population-registry read-db))
 
     (when (.exists discard-dir)
       (dlo/create-metrics-for-dlo! discard-dir))
@@ -259,7 +277,7 @@
                      (throw e)))
           globals {:scf-read-db read-db
                    :scf-write-db write-db
-                   :catalog-hash-debug-dir catalog-hash-debug-dir}]
+                   :pretty-print pretty-print}]
       (transfer-old-messages! (conf/mq-endpoint config))
 
       (when-not disable-update-checking
@@ -294,7 +312,7 @@
 (defprotocol PuppetDBServer
   (shared-globals [this])
   (set-url-prefix [this url-prefix])
-  (query [this query-obj version query-expr paging-options row-callback-fn]
+  (query [this version query-expr paging-options row-callback-fn]
     "Call `row-callback-fn' for matching rows.  The `paging-options' should
     be a map containing :order_by, :offset, and/or :limit."))
 
@@ -306,11 +324,16 @@
   [[:DefaultedConfig get-config]
    [:WebroutingService add-ring-handler get-registered-endpoints]]
   (init [this context]
+
+        (doseq [{:keys [reporter]} (vals metrics/metrics-registries)]
+          (jmx-reporter/start reporter))
         (assoc context :url-prefix (atom nil)))
   (start [this context]
          (start-puppetdb context (get-config) this get-registered-endpoints))
 
   (stop [this context]
+        (doseq [{:keys [reporter]} (vals metrics/metrics-registries)]
+          (jmx-reporter/stop reporter))
         (stop-puppetdb context))
   (set-url-prefix [this url-prefix]
                   (let [old-url-prefix (:url-prefix (service-context this))]
@@ -321,10 +344,12 @@
                               (format "Attempt to set url-prefix to %s when it's already been set to %s" url-prefix @old-url-prefix)))))
   (shared-globals [this]
                   (:shared-globals (service-context this)))
-  (query [this query-obj version query-expr paging-options row-callback-fn]
-         (let [{db :scf-read-db} (get (service-context this) :shared-globals)
-               url-prefix @(get (service-context this) :url-prefix)]
-           (qeng/stream-query-result query-obj version query-expr paging-options db url-prefix row-callback-fn))))
+  (query [this version query-expr paging-options row-callback-fn]
+         (let [sc (service-context this)
+               query-options (-> (get sc :shared-globals)
+                                 (select-keys [:scf-read-db :warn-experimental])
+                                 (assoc :url-prefix @(get sc :url-prefix)))]
+           (qeng/stream-query-result version query-expr paging-options query-options row-callback-fn))))
 
 (def ^{:arglists `([& args])
        :doc "Starts PuppetDB as a service via Trapperkeeper.  Aguments
