@@ -1,24 +1,24 @@
 (ns puppetlabs.puppetdb.query-eng
-  (:require [clojure.java.jdbc :as sql]
+  (:import [org.postgresql.util PGobject])
+  (:require [clojure.core.match :as cm]
+            [clojure.java.jdbc :as sql]
             [clojure.tools.logging :as log]
+            [clojure.set :refer [rename-keys]]
             [puppetlabs.kitchensink.core :as kitchensink]
             [puppetlabs.puppetdb.query.paging :as paging]
             [puppetlabs.puppetdb.cheshire :as json]
             [puppetlabs.puppetdb.http :as http]
             [puppetlabs.puppetdb.jdbc :as jdbc]
             [puppetlabs.puppetdb.query.aggregate-event-counts :as aggregate-event-counts]
-            [puppetlabs.puppetdb.query.catalogs :as catalogs]
             [puppetlabs.puppetdb.query.edges :as edges]
             [puppetlabs.puppetdb.query.events :as events]
             [puppetlabs.puppetdb.query.event-counts :as event-counts]
             [puppetlabs.puppetdb.query.facts :as facts]
-            [puppetlabs.puppetdb.query.factsets :as factsets]
             [puppetlabs.puppetdb.query.fact-contents :as fact-contents]
-            [puppetlabs.puppetdb.query.reports :as reports]
-            [puppetlabs.puppetdb.query.report-data :as report-data]
             [puppetlabs.puppetdb.query.resources :as resources]
-            [puppetlabs.puppetdb.scf.storage-utils :as su]
+            [puppetlabs.puppetdb.scf.storage-utils :as sutils]
             [puppetlabs.puppetdb.schema :as pls]
+            [puppetlabs.puppetdb.utils :as utils]
             [puppetlabs.puppetdb.query-eng.engine :as eng]
             [schema.core :as s]))
 
@@ -36,9 +36,9 @@
                   :rec eng/fact-paths-query}
      :fact-names {:munge facts/munge-name-result-rows
                   :rec eng/fact-names-query}
-     :factsets {:munge factsets/munge-result-rows
+     :factsets {:munge (constantly identity)
                 :rec eng/factsets-query}
-     :catalogs {:munge catalogs/munge-result-rows
+     :catalogs {:munge (constantly identity)
                 :rec eng/catalog-query}
      :nodes {:munge (constantly identity)
              :rec eng/nodes-query}
@@ -48,37 +48,28 @@
               :rec eng/report-events-query}
      :edges {:munge edges/munge-result-rows
              :rec eng/edges-query}
-     :reports {:munge reports/munge-result-rows
+     :reports {:munge (constantly identity)
                :rec eng/reports-query}
-     :report-metrics {:munge (report-data/munge-result-rows :metrics)
+     :report-metrics {:munge (constantly (comp :metrics first))
                       :rec eng/report-metrics-query}
-     :report-logs {:munge (report-data/munge-result-rows :logs)
+     :report-logs {:munge (constantly (comp :logs first))
                    :rec eng/report-logs-query}
      :resources {:munge resources/munge-result-rows
                  :rec eng/resources-query}}))
 
-(defn assoc-in-idx!
-  "assoc into the entity index as if query recs were maps"
-  [[entity & ks] v]
-  (if (and (= :rec (first ks)) (next ks))
-    ;; if it's a query rec and we're updating a nested component, call the
-    ;; existing rec and wrap val in a function
-    (let [rec' (-> ((get-in @entity-fn-idx [entity :rec]))
-                   (assoc-in (next ks) v))]
-      (swap! entity-fn-idx assoc-in [entity :rec] (fn [] rec')))
-    (swap! entity-fn-idx assoc-in (cons entity ks) v)))
-
 (defn get-munge
   [entity]
-  (get-in @entity-fn-idx [entity :munge]))
+  (if-let [munge-result (get-in @entity-fn-idx [entity :munge])]
+    munge-result
+    (throw (IllegalArgumentException. (format "Invalid entity '%s' in query" (utils/dashes->underscores (name entity)))))))
 
 (defn orderable-columns
   [query-rec]
-  (if-not query-rec
-    []
-    (for [[projection-key projection-value] (:projections (query-rec))
-          :when (and (not= projection-key "value") (:queryable? projection-value))]
-      (keyword projection-key))))
+  (vec
+   (for [[projection-key projection-value] (:projections query-rec)
+         :when (and (not= projection-key "value")
+                    (:queryable? projection-value))]
+     (keyword projection-key))))
 
 (defn query->sql
   "Converts a vector-structured `query` to a corresponding SQL query which will
@@ -115,55 +106,92 @@
 
       (munge-fn version url-prefix))))
 
-(defn stream-query-result
+(def query-options-schema
+  {:scf-read-db s/Any
+   :url-prefix String
+   (s/optional-key :warn-experimental) Boolean
+   (s/optional-key :pretty-print) (s/maybe Boolean)})
+
+(pls/defn-validated stream-query-result
   "Given a query, and database connection, return a Ring response with the query
    results."
-  ([entity version query paging-options db url-prefix]
+  ([version query paging-options options]
    ;; We default to doall because tests need this for the most part
-   (stream-query-result entity version query paging-options db url-prefix doall))
-  ([entity version query paging-options db url-prefix row-fn]
-   (let [munge-fn (get-munge-fn entity version paging-options url-prefix)]
-     (jdbc/with-transacted-connection db
+   (stream-query-result version query paging-options options doall))
+  ([version :- s/Keyword
+    query
+    paging-options
+    options :- query-options-schema
+    row-fn]
+   (let [{:keys [scf-read-db url-prefix warn-experimental pretty-print]
+          :or {warn-experimental true
+               pretty-print false}} options
+         {:keys [remaining-query entity]} (eng/parse-query-context version query warn-experimental)
+         munge-fn (get-munge-fn entity version paging-options url-prefix)]
+     (jdbc/with-transacted-connection scf-read-db
        (let [{:keys [results-query]}
-             (query->sql query entity version paging-options)]
+             (query->sql remaining-query entity version paging-options)]
          (jdbc/with-query-results-cursor results-query (comp row-fn munge-fn)))))))
 
-(defn produce-streaming-body
+;; Do we still need this, i.e. do we need the pass-through, and the
+;; strict selectivity in the caller below?
+(defn- coerce-from-json [obj]
+  "Parses obj as JSON if it's a string/stream/reader, otherwise
+  returns obj."
+  (cond
+    (string? obj) (json/parse-strict obj true)
+    (instance? java.io.Reader obj) (json/parse obj true)
+    (instance? java.io.InputStream obj) (json/parse obj true)
+    :else obj))
+
+(pls/defn-validated produce-streaming-body
   "Given a query, and database connection, return a Ring response with
    the query results. query-map is a clojure map of the form
    {:query ['=','certname','foo'] :order_by [{'field':'foo'}]...}
    If the query can't be parsed, a 400 is returned."
-  [entity version query-map db url-prefix]
-  (let [query (:query query-map)
-        query-options (dissoc query-map :query)]
+  [version :- s/Keyword
+   query-map
+   options :- query-options-schema]
+  (let [{:keys [scf-read-db url-prefix warn-experimental pretty-print]
+         :or {warn-experimental true
+              pretty-print false}} options
+        query (:query query-map)
+        {:keys [remaining-query entity paging-clauses]} (eng/parse-query-context
+                                                          version query warn-experimental)
+        paging-options (some-> paging-clauses
+                               (rename-keys {:order-by :order_by})
+                               (update :order_by paging/munge-query-ordering))
+        query-options (merge (dissoc query-map :query) paging-options)]
+
     (try
-      (jdbc/with-transacted-connection db
+      (jdbc/with-transacted-connection scf-read-db
         (let [munge-fn (get-munge-fn entity version query-options url-prefix)
-              {:keys [results-query count-query]} (-> query
-                                                      json/coerce-from-json
+              {:keys [results-query count-query]} (-> remaining-query
+                                                      coerce-from-json
                                                       (query->sql entity version query-options))
               query-error (promise)
               resp (http/streamed-response
-                     buffer
-                     (try (jdbc/with-transacted-connection db
-                            (jdbc/with-query-results-cursor
-                              results-query (comp #(http/stream-json % buffer)
-                                                  #(do (first %) (deliver query-error nil) %)
-                                                  munge-fn)))
-                          (catch java.sql.SQLException e
-                            (deliver query-error e))))]
+                    buffer
+                    (try (jdbc/with-transacted-connection scf-read-db
+                           (jdbc/with-query-results-cursor
+                             results-query (comp #(http/stream-json % buffer pretty-print)
+                                                 #(do (when-not (instance? PGobject %)
+                                                        (first %)) (deliver query-error nil) %)
+                                                 munge-fn)))
+                         (catch java.sql.SQLException e
+                           (deliver query-error e))))]
           (if @query-error
             (throw @query-error)
             (cond-> (http/json-response* resp)
-              count-query (http/add-headers {:count (jdbc/get-result-count count-query)})))))
+                    count-query (http/add-headers {:count (jdbc/get-result-count count-query)})))))
       (catch com.fasterxml.jackson.core.JsonParseException e
-        (log/errorf e (str "Error executing query '%s' for entity '%s' "
-                           "with query-options '%s'. Returning a 400 error code.")
+        (log/errorf e (str "Error executing query '%s' "
+                           "with query options '%s'. Returning a 400 error code.")
                     (name entity) query query-options)
         (http/error-response e))
       (catch IllegalArgumentException e
-        (log/errorf e (str "Error executing query '%s' for entity '%s' "
-                           "with query-options '%s'. Returning a 400 error code.")
+        (log/errorf e (str "Error executing query '%s' "
+                           "with query options '%s'. Returning a 400 error code.")
                     (name entity) query query-options)
         (http/error-response e))
       (catch org.postgresql.util.PSQLException e
@@ -179,15 +207,19 @@
   (let [check-sql (case type
                     :catalog "SELECT 1
                               FROM certnames
+                              INNER JOIN latest_catalogs
+                                ON latest_catalogs.certname_id = certnames.id
                               INNER JOIN catalogs
-                                ON catalogs.certname = certnames.certname
+                                ON catalogs.id = latest_catalogs.certname_id
                               WHERE certnames.certname=?"
+                    :historical-catalog "SELECT 1 FROM catalogs
+                                         WHERE catalog_uuid::text=?"
                     :node "SELECT 1
                            FROM certnames
                            WHERE certname=? "
                     :report (str "SELECT 1
                                   FROM reports
-                                  WHERE " (su/sql-hash-as-str "hash") "=?
+                                  WHERE " (sutils/sql-hash-as-str "hash") "=?
                                   LIMIT 1")
                     :environment "SELECT 1
                                   FROM environments
