@@ -223,16 +223,20 @@
   {:pre  [(kitchensink/datetime? time)]
    :post [(coll? %)]}
   (let [ts (to-timestamp time)]
-    (map :certname (jdbc/query-to-vec "SELECT c.certname FROM certnames c
-                                       LEFT OUTER JOIN catalogs clogs ON c.certname=clogs.certname
-                                       LEFT OUTER JOIN factsets fs ON c.certname=fs.certname
-                                       LEFT OUTER JOIN reports r ON c.certname=r.certname
-                                       WHERE c.deactivated IS NULL
-                                       AND c.expired IS NULL
-                                       AND (clogs.producer_timestamp IS NULL OR clogs.producer_timestamp < ?)
-                                       AND (fs.producer_timestamp IS NULL OR fs.producer_timestamp < ?)
-                                       AND (r.producer_timestamp IS NULL OR r.producer_timestamp < ?)"
-                                      ts ts ts))))
+    (map
+     :certname
+     (jdbc/query-to-vec
+      "select c.certname from certnames c
+         left outer join latest_catalogs lcats on lcats.certname_id = c.id
+         left outer join catalogs cats on cats.id = lcats.catalog_id
+         left outer join factsets fs on c.certname = fs.certname
+         left outer join reports r on c.latest_report_id = r.id
+         where c.deactivated is null
+           and c.expired is null
+           and (cats.producer_timestamp is null or cats.producer_timestamp < ?)
+           and (fs.producer_timestamp is null or fs.producer_timestamp < ?)
+           and (r.producer_timestamp is null or r.producer_timestamp < ?)"
+           ts ts ts))))
 
 (defn node-deactivated-time
   "Returns the time the node specified by `certname` was deactivated, or nil if
@@ -424,15 +428,16 @@
 
 (pls/defn-validated resources-exist? :- #{String}
   "Given a collection of resource-hashes, return the subset that
-  already exist in the database."
+   already exist in the database."
   [resource-hashes :- #{String}]
   (if (seq resource-hashes)
-    (let [query (apply vector
-                       (format "SELECT DISTINCT %s AS resource FROM resource_params_cache WHERE resource %s"
-                               (sutils/sql-hash-as-str "resource")
-                               (jdbc/in-clause resource-hashes))
-                       (map sutils/munge-hash-for-storage resource-hashes))]
-      (jdbc/query-with-resultset query
+    (let [resource-array (->> (vec resource-hashes)
+                              (map sutils/munge-hash-for-storage)
+                              (sutils/array-to-param "bytea" org.postgresql.util.PGobject))
+          query (format "SELECT DISTINCT %s as resource FROM resource_params_cache WHERE resource=ANY(?)"
+                        (sutils/sql-hash-as-str "resource"))
+          sql-params [query resource-array]]
+      (jdbc/query-with-resultset sql-params
                                  #(set (map :resource (sql/result-set-seq %)))))
     #{}))
 
@@ -643,7 +648,6 @@
    edges :- edge-db-schema]
 
   (update! (:catalog-volatility performance-metrics) (count edges))
-  
   (doseq [[source target type] edges]
     ;; This is relatively inefficient. If we have id's for edges, we could do
     ;; this in 1 statement.
@@ -766,27 +770,33 @@
            (let [hash (time! (:catalog-hash performance-metrics)
                              (shash/catalog-similarity-hash catalog))
                  {certname-id :certname_id
+                  stored-hash :catalog_hash
                   latest-producer-timestamp :producer_timestamp} (latest-catalog-metadata certname)]
              (inc! (:updated-catalog performance-metrics))
              (time! (:add-new-catalog performance-metrics)
-                    (let [catalog-id (:id (add-catalog-metadata! hash catalog received-timestamp))]
+                    (time! (get performance-metrics
+                                (if (= stored-hash hash)
+                                  (do (inc! (:duplicate-catalog performance-metrics))
+                                      :catalog-hash-match)
+                                  :catalog-hash-miss))
+                           (let [catalog-id (:id (add-catalog-metadata! hash catalog received-timestamp))]
 
-                      (when-not (some-> latest-producer-timestamp
-                                        (.after (to-timestamp producer_timestamp)))
-                        (let [refs-to-hashes (time! (:resource-hashes performance-metrics)
-                                                    (kitchensink/mapvals shash/resource-identity-hash resources))]
-                          (if-not latest-producer-timestamp
-                            (jdbc/insert! :latest_catalogs {:certname_id certname-id :catalog_id catalog-id})
-                            (jdbc/update! :latest_catalogs {:catalog_id catalog-id} ["certname_id=?" certname-id]))
-                          (update-catalog-associations! certname-id catalog refs-to-hashes)))
-                      (jdbc/delete! :catalogs
-                                    [(str "certname = ? AND "
-                                          "id NOT IN (SELECT id FROM catalogs "
-                                          "           WHERE certname=?"
-                                          "           ORDER BY producer_timestamp DESC LIMIT ?)")
-                                     certname
-                                     certname
-                                     historical-catalogs-limit])))
+                             (when-not (some-> latest-producer-timestamp
+                                               (.after (to-timestamp producer_timestamp)))
+                               (let [refs-to-hashes (time! (:resource-hashes performance-metrics)
+                                                           (kitchensink/mapvals shash/resource-identity-hash resources))]
+                                 (if-not latest-producer-timestamp
+                                   (jdbc/insert! :latest_catalogs {:certname_id certname-id :catalog_id catalog-id})
+                                   (jdbc/update! :latest_catalogs {:catalog_id catalog-id} ["certname_id=?" certname-id]))
+                                 (update-catalog-associations! certname-id catalog refs-to-hashes)))
+                             (jdbc/delete! :catalogs
+                                           [(str "certname = ? AND "
+                                                 "id NOT IN (SELECT id FROM catalogs "
+                                                 "           WHERE certname=?"
+                                                 "           ORDER BY producer_timestamp DESC LIMIT ?)")
+                                            certname
+                                            certname
+                                            historical-catalogs-limit]))))
              hash))))
 
 (pls/defn-validated replace-catalog!
@@ -1229,7 +1239,7 @@
                                    reports/resources->resource-events
                                    (map normalize-resource-event)))))
 
-(pls/defn-validated add-report!*
+(s/defn add-report!*
   "Helper function for adding a report.  Accepts an extra parameter, `update-latest-report?`, which
   is used to determine whether or not the `update-latest-report!` function will be called as part of
   the transaction.  This should always be set to `true`, except during some very specific testing
@@ -1387,7 +1397,7 @@
              (log/warnf "Not updating facts for certname %s because local data is newer." certname))
            (add-facts! fact-data))))
 
-(pls/defn-validated add-report!
+(s/defn add-report!
   "Add a report and all of the associated events to the database."
   [report :- reports/report-wireformat-schema
    received-timestamp :- pls/Timestamp]
